from __future__ import print_function
import os, sys, json, binascii, six, tempfile, zipfile
from ..errors import handle_server_error

APPID = u"lothar.com/wormhole/text-or-file-xfer"

def accept_file(args, them_d, w):
    from ..blocking.transit import TransitReceiver, TransitError
    from .progress import start_progress, update_progress, finish_progress

    file_data = them_d["file"]
    if args.output_file:
        filename = args.output_file
    else:
        # the basename() is intended to protect us against
        # "~/.ssh/authorized_keys" and other attacks
        filename = os.path.basename(file_data["filename"]) # unicode
    filesize = file_data["filesize"]

    # get confirmation from the user before writing to the local directory
    if os.path.exists(filename):
        print("Error: refusing to overwrite existing file %s" % (filename,))
        data = json.dumps({"error": "file already exists"}).encode("utf-8")
        w.send_data(data)
        return 1

    print("Receiving file (%d bytes) into: %s" % (filesize, filename))
    while True and not args.accept_file:
        ok = six.moves.input("ok? (y/n): ")
        if ok.lower().startswith("y"):
            break
        print("transfer rejected", file=sys.stderr)
        data = json.dumps({"error": "transfer rejected"}).encode("utf-8")
        w.send_data(data)
        return 1

    transit_receiver = TransitReceiver(args.transit_helper)
    data = json.dumps({
        "file_ack": "ok",
        "transit": {
            "direct_connection_hints": transit_receiver.get_direct_hints(),
            "relay_connection_hints": transit_receiver.get_relay_hints(),
            },
        }).encode("utf-8")
    w.send_data(data)
    # now done with the Wormhole object

    # now receive the rest of the owl
    tdata = them_d["transit"]
    transit_key = w.derive_key(APPID+u"/transit-key")
    transit_receiver.set_transit_key(transit_key)
    transit_receiver.add_their_direct_hints(tdata["direct_connection_hints"])
    transit_receiver.add_their_relay_hints(tdata["relay_connection_hints"])
    record_pipe = transit_receiver.connect()

    print("Receiving %d bytes for '%s' (%s).." % (filesize, filename,
                                                  transit_receiver.describe()))
    tmp = filename + ".tmp"
    with open(tmp, "wb") as f:
        received = 0
        next_update = start_progress(filesize)
        while received < filesize:
            try:
                plaintext = record_pipe.receive_record()
            except TransitError:
                print()
                print("Connection dropped before full file received")
                print("got %d bytes, wanted %d" % (received, filesize))
                return 1
            f.write(plaintext)
            received += len(plaintext)
            next_update = update_progress(next_update, received, filesize)
        finish_progress(filesize)
        assert received == filesize

    os.rename(tmp, filename)

    print("Received file written to %s" % filename)
    record_pipe.send_record(b"ok\n")
    record_pipe.close()
    return 0

def accept_directory(args, them_d, w):
    from ..blocking.transit import TransitReceiver, TransitError
    from .progress import start_progress, update_progress, finish_progress

    file_data = them_d["directory"]
    mode = file_data["mode"]
    if mode != "zipfile/deflated":
        print("Error: unknown directory-transfer mode '%s'" % (mode,))
        data = json.dumps({"error": "unknown mode"}).encode("utf-8")
        w.send_data(data)
        return 1

    if args.output_file:
        dirname = args.output_file
    else:
        # the basename() is intended to protect us against
        # "~/.ssh/authorized_keys" and other attacks
        dirname = os.path.basename(file_data["dirname"]) # unicode
    filesize = file_data["zipsize"]
    num_files = file_data["numfiles"]
    num_bytes = file_data["numbytes"]

    if os.path.exists(dirname):
        print("Error: refusing to overwrite existing directory %s" % (dirname,))
        data = json.dumps({"error": "directory already exists"}).encode("utf-8")
        w.send_data(data)
        return 1

    print("Receiving directory into: %s/" % (dirname,))
    print("%d files, %d bytes (%d compressed)" % (num_files, num_bytes,
                                                  filesize))
    while True and not args.accept_file:
        ok = six.moves.input("ok? (y/n): ")
        if ok.lower().startswith("y"):
            break
        print("transfer rejected", file=sys.stderr)
        data = json.dumps({"error": "transfer rejected"}).encode("utf-8")
        w.send_data(data)
        return 1

    transit_receiver = TransitReceiver(args.transit_helper)
    data = json.dumps({
        "file_ack": "ok",
        "transit": {
            "direct_connection_hints": transit_receiver.get_direct_hints(),
            "relay_connection_hints": transit_receiver.get_relay_hints(),
            },
        }).encode("utf-8")
    w.send_data(data)
    # now done with the Wormhole object

    # now receive the rest of the owl
    tdata = them_d["transit"]
    transit_key = w.derive_key(APPID+u"/transit-key")
    transit_receiver.set_transit_key(transit_key)
    transit_receiver.add_their_direct_hints(tdata["direct_connection_hints"])
    transit_receiver.add_their_relay_hints(tdata["relay_connection_hints"])
    record_pipe = transit_receiver.connect()

    print("Receiving %d bytes for '%s' (%s).." % (filesize, dirname,
                                                  transit_receiver.describe()))
    f = tempfile.SpooledTemporaryFile()
    received = 0
    next_update = start_progress(filesize)
    while received < filesize:
        try:
            plaintext = record_pipe.receive_record()
        except TransitError:
            print()
            print("Connection dropped before full file received")
            print("got %d bytes, wanted %d" % (received, filesize))
            return 1
        f.write(plaintext)
        received += len(plaintext)
        next_update = update_progress(next_update, received, filesize)
    finish_progress(filesize)
    assert received == filesize
    print("Unpacking zipfile..")
    with zipfile.ZipFile(f, "r", zipfile.ZIP_DEFLATED) as zf:
        zf.extractall(path=dirname)
        # extractall() appears to offer some protection against malicious
        # pathnames. For example, "/tmp/oops" and "../tmp/oops" both do the
        # same thing as the (safe) "tmp/oops".

    print("Received files written to %s/" % dirname)
    record_pipe.send_record(b"ok\n")
    record_pipe.close()
    return 0

@handle_server_error
def receive(args):
    # we're receiving text, or a file
    from ..blocking.transcribe import Wormhole, WrongPasswordError
    assert isinstance(args.relay_url, type(u""))

    with Wormhole(APPID, args.relay_url) as w:
        if args.zeromode:
            assert not args.code
            args.code = u"0-"
        code = args.code
        if not code:
            code = w.input_code("Enter receive wormhole code: ", args.code_length)
        w.set_code(code)

        if args.verify:
            verifier = binascii.hexlify(w.get_verifier()).decode("ascii")
            print("Verifier %s." % verifier)

        try:
            them_bytes = w.get_data()
        except WrongPasswordError as e:
            print("ERROR: " + e.explain(), file=sys.stderr)
            return 1
        them_d = json.loads(them_bytes.decode("utf-8"))
        if "error" in them_d:
            print("ERROR: " + them_d["error"], file=sys.stderr)
            return 1

        if "message" in them_d:
            # we're receiving a text message
            print(them_d["message"])
            data = json.dumps({"message_ack": "ok"}).encode("utf-8")
            w.send_data(data)
            return 0

        if "error" in them_d:
            print("ERROR: " + data["error"], file=sys.stderr)
            return 1

        if "file" in them_d:
            return accept_file(args, them_d, w)

        if "directory" in them_d:
            return accept_directory(args, them_d, w)

        print("I don't know what they're offering\n")
        print("Offer details:", them_d)
        data = json.dumps({"error": "unknown offer type"}).encode("utf-8")
        w.send_data(data)
        return 1



from __future__ import print_function
from binascii import hexlify
from twisted.trial import unittest
from twisted.internet import protocol, defer, reactor
from twisted.internet.endpoints import clientFromString, connectProtocol
from .common import ServerBase

class Accumulator(protocol.Protocol):
    def __init__(self):
        self.data = b""
        self.count = 0
        self._wait = None
    def waitForBytes(self, more):
        assert self._wait is None
        self.count = more
        self._wait = defer.Deferred()
        self._check_done()
        return self._wait
    def dataReceived(self, data):
        self.data = self.data + data
        self._check_done()
    def _check_done(self):
        if self._wait and len(self.data) >= self.count:
            d = self._wait
            self._wait = None
            d.callback(self)
    def connectionLost(self, why):
        if self._wait:
            self._wait.errback("closed")

class Transit(ServerBase, unittest.TestCase):
    @defer.inlineCallbacks
    def test_basic(self):
        ep = clientFromString(reactor, self.transit)
        a1 = yield connectProtocol(ep, Accumulator())
        a2 = yield connectProtocol(ep, Accumulator())

        token1 = b"\x00"*32
        a1.transport.write(b"please relay " + hexlify(token1) + b"\n")
        a2.transport.write(b"please relay " + hexlify(token1) + b"\n")

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        yield a1.waitForBytes(len(exp))
        self.assertEqual(a1.data, exp)
        s1 = b"data1"
        a1.transport.write(s1)

        exp = b"ok\n"
        yield a2.waitForBytes(len(exp))
        self.assertEqual(a2.data, exp)

        # all data they sent after the handshake should be given to us
        exp = b"ok\n"+s1
        yield a2.waitForBytes(len(exp))
        self.assertEqual(a2.data, exp)

        a1.transport.loseConnection()
        a2.transport.loseConnection()

    @defer.inlineCallbacks
    def test_bad_handshake(self):
        ep = clientFromString(reactor, self.transit)
        a1 = yield connectProtocol(ep, Accumulator())

        token1 = b"\x00"*32
        # the server waits for the exact number of bytes in the expected
        # handshake message. to trigger "bad handshake", we must match.
        a1.transport.write(b"please DELAY " + hexlify(token1) + b"\n")

        exp = b"bad handshake\n"
        yield a1.waitForBytes(len(exp))
        self.assertEqual(a1.data, exp)

        a1.transport.loseConnection()

    @defer.inlineCallbacks
    def test_impatience(self):
        ep = clientFromString(reactor, self.transit)
        a1 = yield connectProtocol(ep, Accumulator())

        token1 = b"\x00"*32
        # sending too many bytes is impatience.
        a1.transport.write(b"please RELAY NOWNOW " + hexlify(token1) + b"\n")

        exp = b"impatient\n"
        yield a1.waitForBytes(len(exp))
        self.assertEqual(a1.data, exp)

        a1.transport.loseConnection()



from __future__ import print_function
import os, time
from collections import defaultdict
from ..database import get_db
from ..errors import UsageError

def abbrev(t):
    if t is None:
        return "-"
    if t > 1.0:
        return "%.3fs" % t
    if t > 1e-3:
        return "%.1fms" % (t*1e3)
    return "%.1fus" % (t*1e6)

def print_event(event):
    started, result, waiting_time, total_time = event
    followthrough = None
    if waiting_time and total_time:
        followthrough = total_time - waiting_time
    print("%s: %-6s total=%7s wait=%7s ft=%7s" %
          (time.ctime(started), result,
           abbrev(total_time),
           abbrev(waiting_time),
           abbrev(followthrough),
          ))

def show_usage(args):
    if not os.path.exists("relay.sqlite"):
        raise UsageError("cannot find relay.sqlite, please run from the server directory")
    oldest = None
    newest = None
    counters = defaultdict(int)
    db = get_db("relay.sqlite")
    c = db.execute("SELECT * FROM `usage` ORDER BY `started` ASC LIMIT ?",
                   (args.n,))
    for row in c.fetchall():
        counters["total"] += 1
        counters[row["result"]] += 1
        if oldest is None or row["started"] < oldest:
            oldest = row["started"]
        if newest is None or row["started"] > newest:
            newest = row["started"]
        event = (row["started"], row["result"],
                 row["waiting_time"], row["total_time"])
        print_event(event)
    total = counters["total"]
    if total:
        print("(most recent started %s ago)" % abbrev(time.time() - newest))
        elapsed = time.time() - oldest
        print("%d events in %s (%.2f per hour)" % (total, abbrev(elapsed),
                                                   (3600 * total / elapsed)))
        print(", ".join(["%s=%d (%d%%)" %
                         (k, counters[k], (100.0 * counters[k] / total))
                         for k in sorted(counters)
                         if k != "total"]))
    return 0

def tail_usage(args):
    if not os.path.exists("relay.sqlite"):
        raise UsageError("cannot find relay.sqlite, please run from the server directory")
    db = get_db("relay.sqlite")
    # we don't seem to have unique row IDs, so this is an inaccurate and
    # inefficient hack
    seen = set()
    while True:
        old = time.time() - 2*60*60
        c = db.execute("SELECT * FROM `usage` WHERE `started` > ?"
                       " ORDER BY `started` ASC", (old,))
        for row in c.fetchall():
            event = (row["started"], row["result"],
                     row["waiting_time"], row["total_time"])
            if event not in seen:
                print_event(event)
                seen.add(event)
        time.sleep(2)
    return 0
